// Windows Header Files:
#include <windows.h>
#include <windowsx.h>
#include <commctrl.h>
#include <Shellapi.h>
#include <Shlwapi.h>

// C RunTime Header Files
#include <stdlib.h>
#include <malloc.h>
#include <memory.h>
#include <tchar.h>
#define GLOBAL_HERE
#include "minitray.h"
#include "MntDlg.h"
#include "resource.h"

#define TRAYICONID	1//				ID number for the Notify Icon
#define SWM_TRAYMSG	WM_APP//		the message ID sent to our window
#define SWM_DLG1 (WM_APP+1)
#define SWM_DLG2 (WM_APP+2)
#define SWM_DLG3 (WM_APP+3)
#define SWM_EXIT (WM_APP+4)

// Get dll version number
static ULONGLONG GetDllVersion(LPCTSTR lpszDllName)
{
    ULONGLONG ullVersion = 0;
	HINSTANCE hinstDll;
    hinstDll = LoadLibrary(lpszDllName);
    if(hinstDll)
    {
        DLLGETVERSIONPROC pDllGetVersion;
        pDllGetVersion = (DLLGETVERSIONPROC)GetProcAddress(hinstDll, "DllGetVersion");
        if(pDllGetVersion)
        {
            DLLVERSIONINFO dvi;
            HRESULT hr;
            ZeroMemory(&dvi, sizeof(dvi));
            dvi.cbSize = sizeof(dvi);
            hr = (*pDllGetVersion)(&dvi);
            if(SUCCEEDED(hr))
				ullVersion = MAKEDLLVERULL(dvi.dwMajorVersion, dvi.dwMinorVersion,0,0);
        }
        FreeLibrary(hinstDll);
    }
    return ullVersion;
}

// Name says it all
void ShowContextMenu(HWND hWnd)
{
	POINT pt;
	GetCursorPos(&pt);
	HMENU hMenu = CreatePopupMenu();
	if(hMenu)
	{
        TCHAR szMenuItemText[MAX_PATH+0x8];

		LoadInternatString(IDS_MOUNTPOPUP_MOUNTEDLIST,szMenuItemText,MAX_PATH);
		InsertMenu(hMenu, -1, MF_BYPOSITION, SWM_DLG1, szMenuItemText);		

		LoadInternatString(IDS_MOUNTPOPUP_MOUNTNEW,szMenuItemText,MAX_PATH);
		InsertMenu(hMenu, -1, MF_BYPOSITION, SWM_DLG2, szMenuItemText);		

   		LoadInternatString(IDS_MOUNTPOPUP_ABOUT,szMenuItemText,MAX_PATH);
		InsertMenu(hMenu, -1, MF_BYPOSITION, SWM_DLG3, szMenuItemText);		

   		LoadInternatString(IDS_MOUNTPOPUP_EXIT,szMenuItemText,MAX_PATH);
		InsertMenu(hMenu, -1, MF_BYPOSITION, SWM_EXIT, szMenuItemText);		

		// note:	must set window to the foreground or the
		//			menu won't disappear when it should
		SetForegroundWindow(hWnd);

		TrackPopupMenu(hMenu, TPM_BOTTOMALIGN,
			pt.x, pt.y, 0, hWnd, NULL );
		DestroyMenu(hMenu);
	}
}

LRESULT CALLBACK DummyWndProc(HWND hWnd,UINT message,WPARAM wParam,LPARAM lParam)
{
	switch(message)
	{
		case WM_INITDIALOG:
			break;
	    case SWM_TRAYMSG:
		    switch(lParam)
		    {/*
		    case WM_LBUTTONDBLCLK:
			    ShowWindow(hWnd, SW_RESTORE);
			    break;
                */
		    case WM_RBUTTONDOWN:
		    case WM_CONTEXTMENU:
			    ShowContextMenu(hWnd);
		    }
		    break;
        case WM_COMMAND:
            {
                if (GET_WM_COMMAND_ID(wParam,lParam)==SWM_EXIT)
                    DestroyWindow(hWnd);

                if (GET_WM_COMMAND_ID(wParam,lParam)==SWM_DLG1)
                    DoDialogMountList(hWnd);

                if (GET_WM_COMMAND_ID(wParam,lParam)==SWM_DLG2)
                    DoDialogMountNewImage(hWnd);

                if (GET_WM_COMMAND_ID(wParam,lParam)==SWM_DLG3)
                    DoDialogAbout(hWnd);
            }
            break;

        case WM_DESTROY:
            OutputDebugString("end");            
            PostQuitMessage(0);
            break;
	}

    return (DefWindowProc(hWnd, message, wParam, lParam));
}

#define DUMMY_CLASS_NAME "DummyFDTray"
int APIENTRY WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
	MSG msg;
    HWND hWnd;
    NOTIFYICONDATA niData;
    {
      WNDCLASS wc;

      ghRes=hInstance;  
	  
      memset(&wc,0,sizeof(WNDCLASS));
      wc.hCursor = LoadCursor(NULL, IDC_ARROW);
      wc.hIcon = LoadIcon(hInstance, "IDI_ICONTRAY");
      wc.hInstance = hInstance;
      wc.lpfnWndProc = DummyWndProc;
      wc.lpszClassName = DUMMY_CLASS_NAME;
      RegisterClass(&wc);
      hWnd=CreateWindow(DUMMY_CLASS_NAME,"",WS_OVERLAPPEDWINDOW,CW_USEDEFAULT,CW_USEDEFAULT,0,0,NULL,NULL,hInstance,NULL);
    }


	// Fill the NOTIFYICONDATA structure and call Shell_NotifyIcon

	// zero the structure - note:	Some Windows funtions require this but
	//								I can't be bothered which ones do and
	//								which ones don't.
	ZeroMemory(&niData,sizeof(NOTIFYICONDATA));

	// get Shell32 version number and set the size of the structure
	//		note:	the MSDN documentation about this is a little
	//				dubious and I'm not at all sure if the method
	//				bellow is correct
	ULONGLONG ullVersion = GetDllVersion(_T("Shell32.dll"));
/*
	if(ullVersion >= MAKEDLLVERULL(5, 0,0,0))
		niData.cbSize = sizeof(NOTIFYICONDATA);
	else niData.cbSize = NOTIFYICONDATA_V2_SIZE;

*/
    niData.cbSize = sizeof(NOTIFYICONDATA); //
	// the ID number can be anything you choose
	niData.uID = TRAYICONID;

	// state which structure members are valid
	niData.uFlags = NIF_ICON | NIF_MESSAGE | NIF_TIP;

	// load the icon
	niData.hIcon = LoadIcon(hInstance, "IDI_ICONTRAY");
        /*(HICON)LoadImage(hInstance,"IDI_ICONTRAY",
		IMAGE_ICON, GetSystemMetrics(SM_CXSMICON),GetSystemMetrics(SM_CYSMICON),
		LR_DEFAULTCOLOR);*/

	// the window to send messages to and the message to send
	//		note:	the message value should be in the
	//				range of WM_APP through 0xBFFF
	niData.hWnd = hWnd;
    niData.uCallbackMessage = SWM_TRAYMSG;

	// tooltip message
    LoadInternatString(IDS_TIPS,niData.szTip,sizeof(niData.szTip)/sizeof(TCHAR));    

	Shell_NotifyIcon(NIM_ADD,&niData);

	// free icon handle
	if(niData.hIcon && DestroyIcon(niData.hIcon))
		niData.hIcon = NULL;


    while (GetMessage(&msg, NULL, 0, 0))
    //if (!TranslateAccelerator(hWndMain, hAccel, &msg))
    {
        //if (NULL == hDlgCurrent || !IsDialogMessage(hDlgCurrent, &msg))
        {
            TranslateMessage(&msg);
            DispatchMessage(&msg);
        }
    }
    Shell_NotifyIcon(NIM_DELETE,&niData);
}
